from PlugIns import PluggableContainer
from SheetProviders import SheetProviderContainer, SheetProvider, AcquiredSheetProvider
from AttributeProviders import AttributeProviderContainer, AttributeProvider, AcquiredAttributeProvider
from Globals import HTMLFile, default__class_init__, PersistentMapping
from Globals import Persistent
from OFS.Folder import Folder
from ComputedAttribute import ComputedAttribute
from App.Common import absattr
from PropertySheets import VirtualSheets
from DynPersist import DynPersist
import Products

_marker=[]

SelfKey = 'ZPatterns.Rack','Self'


























class Rack(AttributeProviderContainer, SheetProviderContainer, 
            PluggableContainer):

    """
    Basic Rack - supports sheet and attribute provider plugins, as
    well as persistent storage of any ZClass object and arbitrary
    associated data (using "slots").
    """

    __plugin_groups__ = SheetProviderContainer.__plugin_groups__ + \
                        AttributeProviderContainer.__plugin_groups__


    # Template (i.e. abstract) methods, should be left as-is

    def getItem(self, key):
        """Get an item from the rack by key"""
        item = self.retrieveItem(key)
        if item is not None: item._setRack(self.aq_inner)
        return item

    def newItem(self, key=None):
        """Create a new item - attempts to generate key if not supplied"""
        if key is None:
            key=getattr(self,'newKey',None)
            key=key and key()
            if key is None: return None

        item = self.createItem(key)
        if item is not None: item._setRack(self.aq_inner)
        return item

    def _delClient(self,client):
        """Called by a Rackmountable to delete rack-level data for itself"""

        self._delAllSheetsFor(client)
        self.deleteItem(self._getClientID(client))
        self._delSlotFor(client)



    # Methods to be overridden in subclasses for non-ZODB storage

    def retrieveItem(self,key):
        """Retrieve an object, identified by key"""

        slot = self._readableSlot(key)

        if slot:
            item = slot[SelfKey]
            # XXX class remapping should go here??

            # The next two lines are not needed for non-ZODB storage, so
            # don't copy them if you're subclassing this!
            item._setSlot(slot)
            item._clearPerTransactionCache()

            return item

    def createItem(self,key):
        """Create a new object, identified by key"""

        if self._readableSlot(key):
            raise KeyError,("'%s' already exists" % key)

        slot = self._writeableSlot(key)
        item = self._v_itemConstructor(key)

        slot[SelfKey] = item
        item._setSlot(slot)         # Not needed for non-ZODB storage
        return item


    def deleteItem(self,key):
        # Delete object identified by key - NOT A PUBLIC API!
        # For ZODB-based rack, data is deleted by _delSlotFor call
        # in _delClient().  Non-ZODB racks should do their SQL, LDAP
        # or whatever else kind of deletion here.
        # Note that the proper way to delete a rack-mounted item is to
        # ask it to delete itself!
        pass 

    # Optionally overrideable methods for additional functionality

    #def newKey(self):
    #    """Generate a new object key"""
    #    pass    # must be defined if key autogeneration is desired
    #
    # XXX
    # def countItems(self):
    #    pass
    #
    # def deleteAll(self):
    #    pass


    _initialized = 0
    
    def _setup(self):
        # set up default providers
        
        if self._initialized:
            # We must only initialize once, since _setup gets called on pastes,
            # imports, etc.
            return
            
        self._installPlugIn(
            AcquiredSheetProvider('AcquiredSheets', title='Default Acquired Sheets')
        )
        
        self._installPlugIn(
            SheetProvider('PersistentSheets', title='Default Persistent Sheets')
        )
        
        self._installPlugIn(
            AcquiredAttributeProvider('AcquiredAttributes',
                title='Default Acquired Attributes')
        )
        
        self._installPlugIn(
            AttributeProvider('PersistentAttributes',
                title='Default Persistent Attributes')
        )
        
        self._initialized = 1
        






    # ZClass Machinery (Optionally overrideable)

    _defaultClass = None

    def _zclassOK(self, z):
        """Verify that the ZClass selected is acceptable for this kind of rack"""
        # All we check here is rack-mountability
        return getattr(z._zclass_,'_isRackMountable',0)

    def _v_itemConstructor(self):
        c = getattr(self, '_zclass', None) or self._defaultClass

        if c:
            if type(c) is type(''):
                c = Products.meta_classes[c]
        else:
            def err(key):
                raise TypeError, "No ZClass set - please use 'Storage' tab"
            return err

        c = getattr(c,'_zclass_',c)
        self._v_itemConstructor = c
        return c

    _v_itemConstructor = ComputedAttribute(_v_itemConstructor)


    def _RawItem(self, key):
        """
        Create an empty object of the right class
        (DEPRECATED - use self._v_itemConstructor(key) instead!)
        """
        return self._v_itemConstructor(key)








    def _unifiedZClassRegistry(self):
        """Return a sanely unified registry of ZClasses and ZClass bases"""

        z = Products.meta_classes
        i = Products.meta_class_info

        r = {}
        for k in z.keys():
            r[k] = i[k], z[k]

        for data in self.aq_acquire('_getProductRegistryData')('zclasses'):
            path = '%(product)s/%(id)s' % data
            r[path] = '%(product)s: %(id)s' % data, data['meta_class']

        for k,(i,z) in r.items():
            if not self._zclassOK(z): del r[k]

        return r


    def manage_product_zclass_info(self):
        """Return a list of ZClasses suitable for use in this Rack"""

        Z = self._v_itemConstructor
        Z = getattr(Z, 'aq_self', Z)

        return map(lambda x, Z=Z: {
                'name':x[1][0],
                'path':x[0],
                'selected': (x[1][1]._zclass_ is Z) and 'selected' or '',
            }, 
            self._unifiedZClassRegistry().items()
        )








    # Persistence Machinery - little need to override, but to be extended
    # for future multi-ZODB support

    storageInfo = ''    # Reserved for configuring future multi-ZODB support
    __storage = None    # This rack's storage object for sheets and/or objects

    def storageInUse(self):
        """How many racked items have persistent storage use?"""
        return len(self.__readableStorage)

    def __writeableStorage(self):
        """Create new a mapping-like object for storing 'persistent' stuff and save it in self"""
        from BTree import BTree
        s = self.__readableStorage = self.__writeableStorage = BTree()
        return s
        
    __readableStorage = {}
    __writeableStorage = ComputedAttribute(__writeableStorage)
        

    def _migrateStorage(self,storage):
        """Move existing data to new kind of storage"""

        #if self.storageInUse():
        #    raise ValueError, "Cannot change storage with contents present"

        raise "NotImplemented", "Only self-storage is currently supported"

        # XXX self.storageInfo = storage; del self.__storage

    def manage_storageOptions(self):
        """Return a list of key-value pair tuples listing storage options"""
        return [('', 'Directly in this object')]

    def _getClientID(self,client):
        return absattr(client.id)





    def _readableSlotFor(self, client):
        slot = self._readableSlot(self._getClientID(client),_marker)
        if slot is not _marker: 
            client._setSlot(slot)
            return slot
        else:
            return {}

    def _writeableSlotFor(self, client):
        slot = self._writeableSlot(self._getClientID(client))
        client._setSlot(slot)
        return slot

    def _readableSlot(self, key, default={}):
        return self.__readableStorage.get(key,default)

    def _writeableSlot(self, key):
        r = self.__writeableStorage.get(key,_marker)
        if r is _marker:
            r = self.__writeableStorage[key] = PersistentMapping()
        return r


    def _delSlotFor(self, client):

        """Delete persistent mapping for client, if present"""

        id = self._getClientID(client)
        if self.__readableStorage.has_key(id): del self.__writeableStorage[id]
        











    # Management interface

    manage_storageForm = HTMLFile('www/storageForm', globals())

    def manage_setStorage(self, zclass=None, storage=None, REQUEST=None):
        """
        Change class of returned items, and persistent storage used
        to store them
        """
        if zclass is not None:
            if self._unifiedZClassRegistry().has_key(zclass):
                if Products.meta_classes.has_key(zclass):
                    self._zclass = zclass
                else:
                    self._zclass = self._unifiedZClassRegistry()[zclass][1]
                if hasattr(self,'_v_itemConstructor'):
                    del self._v_itemConstructor
            else:
                raise NameError,("Invalid/nonexistent ZClass '%s'" % zclass)

        if storage is not None and storage != self.storageInfo:
            self._migrateStorage(storage)

        if REQUEST is not None:
            return self.manage_storageForm(self, REQUEST,
                manage_tabs_message='Updated storage settings.')

    # XXX
    # def manage_emptyRack(self, REQUEST=None):
    #    """Wipe all items from storage"""
    #    self.deleteAll()    # XXX delete actual storage too?
    #    if REQUEST is not None:
    #        return self.manage_main(self, REQUEST, update_menu=1,
    #            message='Removed all items.')

    manage_options_right = (
                {'label':'Storage',
                 'action':'manage_storageForm'},
    ) + PluggableContainer.manage_options_right


    meta_type = "Rack"
    __plugin_kind__ = "Rack"
    icon = 'misc_/ZPatterns/rack'

    __ac_permissions__ = (
        ('Change Rack storage settings',
            ('manage_storageForm','manage_setStorage',
             'manage_product_zclass_info','manage_storageOptions',
            ),
        ),

        ('Access contents information',
            ('getItem','retrieveItem','storageInUse'), # 'countItems'
        ),

        ('Add Items to Racks', ('newItem','createItem','newKey'), ),
    )


default__class_init__(Rack)

manage_addRackForm = HTMLFile('addRack', globals())

def manage_addRack(self, id, title='', REQUEST=None):
    """Add a Rack"""
    ob = Rack(id, title)
    return self.Destination()._installPlugIn(ob,'Added Rack.',REQUEST)














_v_rack = '_v_rack'

class RackMountable(DynPersist,Persistent):
    """
    Mixin for classes that want to delegate storage issues to their Rack
    """

    _isRackMountable = 1
       
    def __init__(self, id):
        # Bypass attribute providers for setting up initial id, 
        # since they need the id in order to work!
        self.__dict__['id']=id; self._p_changed = 1
        
    def _setRack(self, source):
        """Called by the rack before returning the object"""
        self.__dict__[_v_rack] = source


    def _setSlot(self, slot):
        """
        Set up the slot object for persistent storage of object
        'baggage' (e.g. WebDAV property sheets, etc.).  Called by
        ZODB-based racks at object retrieval, or by non-ZODB based
        racks at the object's request.
        """
        self._v_readableSlot = self._v_writeableSlot = slot














    # Computed attributes to handle on-demand slot retrieval/allocation

    __per_transaction_cache_attrs__ = '_v_currentSheets', '_v_attrCache'

    def _clearPerTransactionCache(self):
        d=self.__dict__; have=d.has_key
        for a in self.__per_transaction_cache_attrs__:
            if have(a): del d[a]

    def _v_readableSlot(self):
        return self.__dict__[_v_rack]._readableSlotFor(self)

    def _v_writeableSlot(self):
        return self.__dict__[_v_rack]._writeableSlotFor(self)

    def _v_currentSheets(self):
        self._v_currentSheets = tuple(
            self.__dict__[_v_rack]._PropertySheetsFor(self)
        )
        return self._v_currentSheets

    def _v_attrCache(self):
        v = self._v_attrCache = {}
        return v

    _v_readableSlot = ComputedAttribute(_v_readableSlot)
    _v_writeableSlot = ComputedAttribute(_v_writeableSlot)
    _v_currentSheets = ComputedAttribute(_v_currentSheets)
    _v_attrCache = ComputedAttribute(_v_attrCache)












    def __get_attr__(self,name,_v_rack=_v_rack):

        cache = self._v_attrCache

        v = cache.get(name,_marker)
        if v is not _marker:
            return v

        v = self.__dict__[_v_rack]._AttributeFor(self,name,_marker)
        if v is not _marker: 
            cache[name]=v
            return v

        raise AttributeError, name


    def __set_attr__(self,name,val,_v_rack=_v_rack):
        v = self._v_attrCache
        if v.has_key(name):
            del v[name]
        
        if not self.__dict__[_v_rack]._SetAttributeFor(self,name,val):
            raise AttributeError, name


    def __del_attr__(self,name,_v_rack=_v_rack):
        v = self._v_attrCache
        if v.has_key(name):
            del v[name]

        if not self.__dict__[_v_rack]._DelAttributeFor(self,name,val):
            raise AttributeError, name









    propertysheets = VirtualSheets()

    def _addPropertySheet(self,name='',xmlns='',_v_rack=_v_rack):
        """
        Return new propertysheet matching name or xmlns, if possible.
        Otherwise return None.
        """
        d=self.__dict__
        if d.has_key('_v_currentSheets'):
            del d['_v_currentSheets']
        return d[_v_rack]._addPropertySheetFor(self,name,xmlns)


    def _delPropertySheet(self,name='',xmlns='',_v_rack=_v_rack):
        """
        Delete propertysheet matching name or xmlns, if present.  Always
        returns None (unless an unexpected error occurs, of course)
        """
        d=self.__dict__
        if d.has_key('_v_currentSheets'):
            del d['_v_currentSheets']
        d[_v_rack]._delPropertySheetFor(self,name,xmlns)


    def manage_delete(self,_v_rack=_v_rack):
        """Delete this object"""
        self.__dict__[_v_rack]._delClient(self)
        return 'OK'

    meta_type = "Rack-mountable"

    __ac_permissions__ = (
        ('Delete rack-mounted objects', ('manage_delete',)),
    )







    # DEPRECATED STUFF 

    def PropertySheets(self):
        """Returns a 'propertysheets'-type object - DEPRECATED"""
        return self.propertysheets


    def PropertySheet(self, name='', xmlns=''):
        """Returns a specific property sheet"""
        return self.__dict__[_v_rack]._PropertySheetFor(self, name, xmlns)




default__class_init__(RackMountable)


















